/*
**
* BEGIN_COPYRIGHT
*
* Copyright (C) 2008-2018 SciDB, Inc.
* All Rights Reserved.
*
* SciDB is free software: you can redistribute it and/or modify
* it under the terms of the AFFERO GNU General Public License as published by
* the Free Software Foundation.
*
* SciDB is distributed "AS-IS" AND WITHOUT ANY WARRANTY OF ANY KIND,
* INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,
* NON-INFRINGEMENT, OR FITNESS FOR A PARTICULAR PURPOSE. See
* the AFFERO GNU General Public License for the complete license terms.
*
* You should have received a copy of the AFFERO GNU General Public License
* along with SciDB.  If not, see <http://www.gnu.org/licenses/agpl-3.0.html>
*
* END_COPYRIGHT
*/

/*
 * @file Operator.cpp
 *
 * @author roman.simakov@gmail.com
 *
 * @brief Implementation of basic operator methods.
 */

#include <memory>
#include <boost/function.hpp>
#include <boost/bind.hpp>
#include <unordered_set>
#include <log4cxx/logger.h>

#include <query/Aggregate.h>
#include <query/QueryPlanUtilites.h>
#include <query/Operator.h>
#include <query/OperatorLibrary.h>
#include <network/NetworkManager.h>
#include <network/MessageDesc.h>
#include <network/MessageUtils.h>
#include <network/NetworkManager.h>
#include <system/SystemCatalog.h>
#include <array/DBArray.h>
#include <array/TransientCache.h>
#include <query/QueryProcessor.h>
#include <system/BlockCyclic.h>
#include <system/Config.h>
#include <system/SciDBConfigOptions.h>
#include <boost/functional/hash.hpp>
#include <util/Hashing.h>
#include <util/Timing.h>
#include <util/MultiConstIterators.h>
#include <rbac/Rights.h>
#include <rbac/Session.h>

using namespace std;

namespace scidb
{

// Logger for operator. static to prevent visibility of variable outside of file
static log4cxx::LoggerPtr logger(log4cxx::Logger::getLogger("scidb.qproc.operator"));

template <typename T>
inline static T min( T const& lhs, T const& rhs)
{
    return lhs < rhs ? lhs : rhs;
}

template <typename T>
inline static T max( T const& lhs, T const& rhs)
{
    return lhs > rhs ? lhs : rhs;
}


#define stringify(name) #name

static const char *OperatorParamPlaceholderTypeNames[] =
{
        stringify(PLACEHOLDER_INPUT),
        stringify(PLACEHOLDER_ARRAY_NAME),
        stringify(PLACEHOLDER_ATTRIBUTE_NAME),
        stringify(PLACEHOLDER_DIMENSION_NAME),
        stringify(PLACEHOLDER_CONSTANT),
        stringify(PLACEHOLDER_EXPRESSION),
        stringify(PLACEHOLDER_VARIES),
        stringify(PLACEHOLDER_SCHEMA),
        stringify(PLACEHOLDER_AGGREGATE_CALL),
        stringify(PLACEHOLDER_NS_NAME),
        stringify(PLACEHOLDER_END_OF_VARIES)
};

void OperatorParamPlaceholder::toString(std::ostream &out, int indent) const
{
    Indent prefix(indent);
    out << prefix(' ', false);
    int typeIndex = 0, type = _placeholderType;
    while ((type >>= 1) != 0) {
        typeIndex += 1;
    }
    out << "[opParamPlaceholder] " << OperatorParamPlaceholderTypeNames[typeIndex];
    out << " requiredType " <<_requiredType.name();
    out <<" ischeme "<<_inputSchema<<"\n";
}

void OperatorParam::toString(std::ostream &out, int indent) const
{
    Indent prefix(indent);
    out << prefix(' ', false);
    out <<"[param] type "<<_paramType<<"\n";
}

void OperatorParamReference::toString(std::ostream &out, int /*indent*/) const
{
    out << "object "<<_objectName;
    out << " inputNo " <<_inputNo;
    out << " objectNo " <<_objectNo;
    out << " inputScheme "<<_inputScheme;
    out << "\n";
}

void OperatorParamArrayReference::toString(std::ostream &out, int indent) const
{
    Indent prefix(indent);
    out << prefix(' ', false);
    out << "[paramArrayReference] ";
    OperatorParamReference::toString(out, indent);
}

void OperatorParamAttributeReference::toString(std::ostream &out, int indent) const
{
    Indent prefix(indent);
    out << prefix(' ', false);
    out << "[paramAttributeReference] ";
    OperatorParamReference::toString(out,indent);
}

void OperatorParamDimensionReference::toString(std::ostream &out, int indent) const
{
    Indent prefix(indent);
    out << prefix(' ', false);
    out << "[paramDimensionReference] ";
    OperatorParamReference::toString(out,indent);
}

void OperatorParamLogicalExpression::toString(std::ostream &out, int indent) const
{
    Indent prefix(indent);
    out << prefix(' ', false);
    out << "[paramLogicalExpression] type "<<_expectedType.name();
    out << " const " << _constant << "\n";
    _expression->toString(out, indent+1);
}

void OperatorParamPhysicalExpression::toString(std::ostream &out, int indent) const
{
    Indent prefix(indent);
    out << prefix(' ', false);
    out << "[paramPhysicalExpression] const " << _constant;
    out << "\n";
    _expression->toString(out, indent+1);
}

void OperatorParamSchema::toString(std::ostream &out, int indent) const
{
    Indent prefix(indent);
    out << prefix(' ', false);
    out <<"[paramSchema] " << _schema <<"\n";
}

void OperatorParamAggregateCall::toString(std::ostream &out, int indent) const
{
    Indent prefix(indent);
    out << prefix(' ', false);
    out << "[paramAggregateCall] " << _aggregateName << "\n" ;

    out << prefix(' ', false);
    out <<"input: ";
    _inputAttribute->toString(out);

    if (_alias.size() )
    {
        out << prefix(' ');
        out << "alias " << _alias << "\n";
    }
}

void OperatorParamNamespaceReference::toString(std::ostream &out, int indent) const
{
    Indent prefix(indent);
    out << prefix(' ', false);
    out << "[paramNsName] " << getNsName() << "(id=" << getNsId() << ")\n";
}

void OperatorParamAsterisk::toString(std::ostream &out, int indent) const
{
    Indent prefix(indent);
    out << prefix(' ', false);
    out << "[paramAsterisk] *" << "\n";
}

void LogicalOperator::toString(std::ostream &out, int indent) const
{
    Indent prefix(indent);
    out << prefix('>', false);
    out << "[lOperator] "<<_logicalName;
    out << " ddl "<< _properties.ddl << "\n";

    for (size_t i = 0; i < _parameters.size(); i++)
    {
        _parameters[i]->toString(out, indent+1);
    }

    for ( size_t i = 0; i < _paramPlaceholders.size(); i++)
    {
        _paramPlaceholders[i]->toString(out,indent+1);
    }

    out << prefix('>', false);
    out << "schema: " << _schema << "\n";
}

PhysicalOperator::PhysicalOperator(std::string const& logicalName,
                                   std::string const& physicalName,
                                   Parameters const& parameters,
                                   ArrayDesc const& schema)
:
    _parameters(parameters),
    _schema(schema),
    _tileMode(false),
    _logicalName(logicalName),
    _physicalName(physicalName)
{
}

PhysicalOperator::~PhysicalOperator()
{
    SCIDB_ASSERT(this);               // NOCHECKIN
}

std::shared_ptr<PhysicalOperator> PhysicalOperator::getShared()
{
    SCIDB_ASSERT(this);
    return shared_from_this();
}

std::weak_ptr<PhysicalOperator> PhysicalOperator::getWeak()
{
    SCIDB_ASSERT(this);
    return std::weak_ptr<PhysicalOperator>(shared_from_this()); // TODO when >= C++17, use weak_from_this()
}

void PhysicalOperator::setOperatorContext(std::shared_ptr<OperatorContext> const& opContext,
                                         std::shared_ptr<JobQueue> const& jobQueue)
{
    static const char * funcName = "PhysOp::setOperatorContext()" ;
    SCIDB_ASSERT(opContext);
    SCIDB_ASSERT(!_operatorContext);

    _operatorContext = opContext;

    LOG4CXX_TRACE(logger, funcName << " this " << static_cast<void*>(this)
                                   << " getPhysicalName() " << getPhysicalName()
                                   << " getOperatorID() " << getOperatorID()
                                   << " operatorContext " << static_cast<void*>(_operatorContext.get()));

    if(false){ // TODO: need to eliminate need for operators to
               //       to start the SG queue ... could run
               //       for all query duration
        // NOTE: query->startSGQueue() is still called externally
        // NOTE: sort depends on the current behavior
        assert(_query.lock()->getSGQueue());
        _query.lock()->getSGQueue()->start(jobQueue);
    }
}

std::shared_ptr<OperatorContext> PhysicalOperator::getOperatorContext() const
{
    return _operatorContext;
}

void PhysicalOperator::unsetOperatorContext()
{
    static const char * funcName = "PhysOp::unsetOperatorContext()" ;

    LOG4CXX_TRACE(logger, funcName << " this " << static_cast<void*>(this)
                                   << " getPhysicalName() " << getPhysicalName()
                                   << " getOperatorID() " << getOperatorID()
                                   << " operatorContext " << static_cast<void*>(_operatorContext.get()));

    if (_operatorContext) {
        _operatorContext.reset();
    } else {
        LOG4CXX_TRACE(logger, funcName << " returning early, nothing to unset");
    }

    if(false){ // TODO: need to eliminate need to stop the SG queue
               //       see setOperatorContext
        // NOTE: query->stopSGQueue() still called externally
        // NOTE: sort might depend on the current behavior
        assert(_query.lock()->getSGQueue());
        _query.lock()->getSGQueue()->stop();
    }
}

void PhysicalOperator::setQuery(const std::shared_ptr<Query>& query)
{
    auto myQuery = _query.lock();
    if(!myQuery) {  // have never registered myself for message lookup
        LOG4CXX_TRACE(logger, "PhysOp::setQuery(): new getOperatorID() " << getOperatorID() );
        query->registerPhysicalOperator(getShared());

        arena::Options options;                              // Arena ctor args
        options.name  (_physicalName.c_str());               // Use operator name
        options.parent(query->getArena());                   // Attach to query
        options.threading(!isSingleThreaded());              // the op may be multi-threaded
        _arena = arena::newArena(options);                   // Create new Arena

        _query = query; // note that setQuery has now been called once
    } else {
        SCIDB_ASSERT(myQuery.get() == query.get()); // must match on subsequent calls
    }
}

OperatorID PhysicalOperator::getOperatorID() const
{
    return _operatorID ;
}

void PhysicalOperator::setOperatorID(const OperatorID& opID)
{
    _operatorID = opID;
}

void PhysicalOperator::toString(std::ostream &out, int indent) const
{
    Indent prefix(indent);
    out << prefix(' ', false);
    out << "schema " <<_schema<<"\n";
}

void PhysicalOperator::dumpArrayToLog(std::shared_ptr<Array> const& input, log4cxx::LoggerPtr& logger)
{
    ArrayDesc const& schema = input->getArrayDesc();
    Attributes const& attrs = schema.getAttributes(true);
    AttributeID const nAttrs = safe_static_cast<AttributeID>(attrs.size());
    vector<FunctionPointer> converters(nAttrs,NULL);
    FunctionLibrary *functionLib = FunctionLibrary::getInstance();
    vector<std::shared_ptr<ConstArrayIterator> > aiters (nAttrs);
    vector<std::shared_ptr<ConstChunkIterator> > citers (nAttrs);
    for (AttributeID i =0; i<nAttrs; ++i)
    {
        TypeId const& typeId = attrs[i].getType();
        converters[i] = functionLib->findConverter(typeId, TID_STRING, false, false, NULL);
        aiters[i] = input->getConstIterator(i);
    }
    while (!aiters[0]->end())
    {
        for(size_t i=0; i<nAttrs; ++i)
        {
            citers[i] = aiters[i]->getChunk().getConstIterator(ChunkIterator::IGNORE_EMPTY_CELLS);
        }
        while (!citers[0]->end())
        {
            Coordinates const& position = citers[0]->getPosition();
            ostringstream out;
            out<<CoordsToStr(position)<<" ";
            for(size_t i=0; i<nAttrs; ++i)
            {
                Value const& v = citers[i]->getItem();
                if (v.isNull())
                {
                    if(v.getMissingReason() == 0)
                    {
                        out<<"[null]";
                    }
                    else
                    {
                        out<<"[?"<<v.getMissingReason()<<"]";
                    }
                }
                else if (converters[i])
                {
                    Value const* input = &v;
                    Value result;
                    converters[i](&input, &result, NULL);
                    out<<result.getString();
                }
                else
                {
                    out<<"[nct]";
                }
                out<<",";
            }
            LOG4CXX_DEBUG(logger, out.str());
            for(size_t i=0; i<nAttrs; ++i)
            {
                ++(*citers[i]);
            }
        }
        for(size_t i=0; i<nAttrs; ++i)
        {
            ++(*aiters[i]);
        }
    }
}

std::shared_ptr<Array>
PhysicalOperator::ensureRandomAccess(std::shared_ptr<Array>& input,
                                     std::shared_ptr<Query> const& query)
{
    if (input->getSupportedAccess() != Array::RANDOM)
    {
        LOG4CXX_DEBUG(logger, "Query "<<query->getQueryID()<<
                      " materializing input "<<input->getArrayDesc());
        auto memArr = std::make_shared<MemArray>(input->getArrayDesc(), query);
        SCIDB_ASSERT(memArr);
        memArr->appendHorizontal(input);
        input = memArr;
    }

    return input;
}

RedistributeContext
PhysicalOperator::getOutputDistribution(std::vector<RedistributeContext> const& inputDistributions,
                                        std::vector<ArrayDesc> const& inputSchemas) const
{
    /// XXX TODO: we should probably get rid of changesDistribution() altogether
    if (changesDistribution(inputSchemas)) {
        throw USER_EXCEPTION(SCIDB_SE_INTERNAL,
                             SCIDB_LE_NOT_IMPLEMENTED)
        << "getOutputDistribution";
    }

    if (inputDistributions.empty()) {

        std::shared_ptr<Query> query(_query);
        SCIDB_ASSERT(query);
        return RedistributeContext(defaultPartitioning(),
                                   query->getDefaultArrayResidency());
    } else {

        ArrayDesc* mySchema = const_cast<ArrayDesc*>(&_schema);
        mySchema->setResidency(inputDistributions[0].getArrayResidency());
        ArrayDistPtr dist = convertToDefaultRedundancy(inputDistributions[0].getArrayDistribution());
        mySchema->setDistribution(dist);

        return RedistributeContext(mySchema->getDistribution(),
                                   mySchema->getResidency());

    }
}

ArrayDistPtr
PhysicalOperator::convertToDefaultRedundancy(ArrayDistPtr const& inputDist)
{
    if ( inputDist->getRedundancy() != DEFAULT_REDUNDANCY) {
        size_t instanceShift(0);
        Coordinates offset;
        ArrayDistributionFactory::getTranslationInfo(inputDist.get(), offset, instanceShift);

        std::shared_ptr<CoordinateTranslator> translator =
        ArrayDistributionFactory::createOffsetTranslator(offset);

        ArrayDistPtr dist =
        ArrayDistributionFactory::getInstance()->construct(inputDist->getPartitioningSchema(),
                                                           DEFAULT_REDUNDANCY,
                                                           inputDist->getContext(),
                                                           translator,
                                                           instanceShift);
        return dist;
    }
    return inputDist;
}

/*
 * PhysicalBoundaries methods
 */
PhysicalBoundaries::PhysicalBoundaries(Coordinates const& start, Coordinates const& end, double density):
    _startCoords(start), _endCoords(end), _density(density)
{
    if (_startCoords.size() != _endCoords.size()) {
        throw SYSTEM_EXCEPTION(SCIDB_SE_INTERNAL, SCIDB_LE_MISMATCHED_COORDINATES_IN_PHYSICAL_BOUNDARIES)
            << "number of dimensions";
    }

    if (density < 0.0 || density > 1.0) {
        throw SYSTEM_EXCEPTION(SCIDB_SE_INTERNAL, SCIDB_LE_MISMATCHED_COORDINATES_IN_PHYSICAL_BOUNDARIES)
            << "density not between 0.0 and 1.0";
    }

    for (size_t i = 0; i< _startCoords.size(); i++)
    {
        if (_startCoords[i] < CoordinateBounds::getMin())
        {
            _startCoords[i] = CoordinateBounds::getMin();
        }
        else if (_startCoords[i] > CoordinateBounds::getMax())
        {
            _startCoords[i] = CoordinateBounds::getMax();
        }

        if (_endCoords[i] < CoordinateBounds::getMin())
        {
            _endCoords[i] = CoordinateBounds::getMin();
        }
        else if (_endCoords[i] > CoordinateBounds::getMax())
        {
            _endCoords[i] = CoordinateBounds::getMax();
        }
    }
}

PhysicalBoundaries PhysicalBoundaries::createFromFullSchema(ArrayDesc const& schema )
{
    Coordinates resultStart, resultEnd;

    for (size_t i =0; i<schema.getDimensions().size(); i++)
    {
        resultStart.push_back(schema.getDimensions()[i].getStartMin());
        resultEnd.push_back(schema.getDimensions()[i].getEndMax());
    }

    return PhysicalBoundaries(resultStart, resultEnd);
}

PhysicalBoundaries PhysicalBoundaries::createEmpty(size_t numDimensions)
{
    Coordinates resultStart, resultEnd;

    for (size_t i =0; i<numDimensions; i++)
    {
        resultStart.push_back(CoordinateBounds::getMax());
        resultEnd.push_back(CoordinateBounds::getMin());
    }

    return PhysicalBoundaries(resultStart, resultEnd);
}

bool PhysicalBoundaries::isEmpty() const
{
    if (_startCoords.size() == 0)
    {
        return true;
    }

    for (size_t i = 0; i<_startCoords.size(); i++)
    {
        if (_startCoords[i] > _endCoords[i])
        {
            return true;
        }
    }
    return false;
}

uint64_t PhysicalBoundaries::getCellNumber (Coordinates const& coords, Dimensions const& dims)
{
    if (dims.size() != coords.size())
    {
        throw SYSTEM_EXCEPTION(SCIDB_SE_INTERNAL, SCIDB_LE_MISMATCHED_COORDINATES_IN_PHYSICAL_BOUNDARIES)
            << "number of dimensions";
    }
    uint64_t result = 0;
    for ( size_t i = 0, n = dims.size(); i < n; i++)
    {
        if (dims[i].getLength()==0)
        {
            throw SYSTEM_EXCEPTION(SCIDB_SE_INTERNAL, SCIDB_LE_MISMATCHED_COORDINATES_IN_PHYSICAL_BOUNDARIES)
                << "dimension has zero length";
        }
        uint64_t t = result * dims[i].getLength();
        if (t / dims[i].getLength() != result) //overflow check multiplication
        {
            return CoordinateBounds::getMaxLength();
        }
        result = t;
        t = result + coords[i] - dims[i].getStartMin();
        if (t < result) //overflow check addition
        {
            return CoordinateBounds::getMaxLength();
        }
        result = t;
    }
    if (!CoordinateBounds::isValidLength(result))
    {
        return CoordinateBounds::getMaxLength();
    }
    return result;
}

Coordinates PhysicalBoundaries::getCoordinates(uint64_t origCellNum, Dimensions const& dims, bool throwOnOverflow)
{
    uint64_t cellNum = origCellNum;  // working copy
    if(cellNum >= CoordinateBounds::getMaxLength())
    {
        return Coordinates(dims.size(), CoordinateBounds::getMax());
    }
    Coordinates result (dims.size(), 0);
    for (int i = safe_static_cast<int>(dims.size()); --i >= 0;)
    {
        result[i] = dims[i].getStartMin() + (cellNum % dims[i].getLength());
        cellNum /= dims[i].getLength();
    }
    if (cellNum != 0)
    {
        if (throwOnOverflow)
        {
            stringstream ss;
            ss << "Non-zero remainder " << cellNum << " for cell number " << origCellNum << " in " << dims;
            throw SYSTEM_EXCEPTION(SCIDB_SE_INTERNAL, SCIDB_LE_MISMATCHED_COORDINATES_IN_PHYSICAL_BOUNDARIES)
                << ss.str();
        }
        else
        {
            LOG4CXX_DEBUG(logger, "Non-zero remainder " << cellNum << " for cell number " << origCellNum
                          << " in " << dims << ", using {*,...}");
            return Coordinates(dims.size(), CoordinateBounds::getMax());
        }
    }
    return result;
}

Coordinates PhysicalBoundaries::reshapeCoordinates (Coordinates const& inCoords,
                                                    Dimensions const& currentDims,
                                                    Dimensions const& newDims,
                                                    bool strict)
{
    uint64_t cellNum = getCellNumber(inCoords, currentDims);
    if ( cellNum >= CoordinateBounds::getMaxLength() )
    {
        return Coordinates(newDims.size(), CoordinateBounds::getMax());
    }
    return getCoordinates(cellNum, newDims, strict);
}

uint64_t PhysicalBoundaries::getNumCells (Coordinates const& start, Coordinates const& end)
{
    if (PhysicalBoundaries(start,end).isEmpty())
    {
        return 0;
    }
    uint64_t result = 1;
    for ( size_t i = 0; i < end.size(); i++)
    {
        ASSERT_EXCEPTION(start[i] >= CoordinateBounds::getMin(),
                         "Invalid start[i], smaller than minimum coordinate");
        ASSERT_EXCEPTION(end[i] <= CoordinateBounds::getMax(),
                         "Invalid end[i], bigger than maximum coordinate");
        if(end[i] >= start[i])
        {
            uint64_t t = result * (end[i] - start[i] + 1);
            if ( t / (end[i] - start[i] + 1) != result) //overflow check multiplication
            {
                return CoordinateBounds::getMaxLength();
            }
            result = t;
        }
        else
        {
            result *= 0;
        }
    }
    if (!CoordinateBounds::isValidLength(result))
    {
        return CoordinateBounds::getMaxLength();
    }
    return result;
}

uint64_t PhysicalBoundaries::getNumCells() const
{
    return getNumCells(_startCoords, _endCoords);
}

uint64_t PhysicalBoundaries::getNumChunks(Dimensions const& dims) const
{
    if (_startCoords.size() != dims.size())
    {
        throw SYSTEM_EXCEPTION(SCIDB_SE_INTERNAL, SCIDB_LE_MISMATCHED_COORDINATES_IN_PHYSICAL_BOUNDARIES)
            << "number of dimensions";
    }

    if (isEmpty())
    {
        return 0;
    }

    uint64_t result = 1;
    for (size_t i =0; i < _endCoords.size(); i++)
    {
        if (_startCoords[i]<=CoordinateBounds::getMin() || _endCoords[i]>=CoordinateBounds::getMax())
        {
            return CoordinateBounds::getMaxLength();
        }

        DimensionDesc const& dim = dims[i];
        if (!dim.isIntervalResolved() )
        {
            throw UnknownChunkIntervalException(__FILE__, __LINE__);
        }

        if (_startCoords[i] < dim.getStartMin() || _endCoords[i] > dim.getEndMax())
        {
            stringstream ss;
            ss << "target dimension too small";
            if (_startCoords[i] < dim.getStartMin()) {
                ss << std::endl << "----- _startCoords=" << _startCoords[i] << " is less than " << "getStartMin=" << dim.getStartMin();
            }

            if (_endCoords[i] > dim.getEndMax()) {
                ss << std::endl << "----- _endCoords=" << _endCoords[i] << " is greater than " << "getEndMax=" << dim.getEndMax();
            }

            // Note: *not* an internal error, this can happen for example if we insert(unbounded, bounded)
            // and the data in the unbounded (i.e. <...>[i=0:*,...]) array does not fit inside the bounded one.
            // See ticket #2016.
            throw SYSTEM_EXCEPTION(SCIDB_SE_OPERATOR, SCIDB_LE_MISMATCHED_COORDINATES_IN_PHYSICAL_BOUNDARIES)
                << ss.str();
        }

        Coordinate arrayStart = dim.getStartMin();
        int64_t chunkInterval = dim.getChunkInterval();
        Coordinate physStart = _startCoords[i]; //TODO:OPTAPI (- overlap) ?
        Coordinate physEnd = _endCoords[i];
        int64_t numChunks = chunkInterval == 0 ? 0 :
            ((physEnd - arrayStart + chunkInterval) / chunkInterval) - ((physStart - arrayStart) / chunkInterval);

        uint64_t t = result * numChunks;
        if ( numChunks && t / numChunks != result) //overflow check multiplication
        {
            return CoordinateBounds::getMaxLength();
        }
        result = t;
    }
    return result;
}

std::ostream& operator<< (std::ostream& stream, const PhysicalBoundaries& bounds)
{
    stream<<"start "<<bounds._startCoords<<" end "<<bounds._endCoords<<" density "<<bounds._density;
    return stream;
}

bool PhysicalBoundaries::isInsideBox (Coordinate const& in, size_t const& dimensionNum) const
{
    if (dimensionNum >= _startCoords.size())
        throw SYSTEM_EXCEPTION(SCIDB_SE_INTERNAL, SCIDB_LE_WRONG_NUMBER_OF_DIMENSIONS);

    Coordinate start = _startCoords[dimensionNum];
    Coordinate end = _endCoords[dimensionNum];

    return in >= start && in <= end;
}

bool PhysicalBoundaries::isInsideBox (Coordinates const& in) const
{
    if ( in.size() > _startCoords.size())
        throw SYSTEM_EXCEPTION(SCIDB_SE_INTERNAL, SCIDB_LE_WRONG_NUMBER_OF_DIMENSIONS);

    for ( size_t i = 0, len = in.size(); i < len; ++i )
    {
        if (! isInsideBox ( in[i], i ) )
        {
          return false;
        }
    }

    return true;
}


PhysicalBoundaries PhysicalBoundaries::intersectWith (PhysicalBoundaries const& other) const
{
    if (_startCoords.size() != other._startCoords.size()
            || _startCoords.size() != other._endCoords.size())
        throw SYSTEM_EXCEPTION(SCIDB_SE_INTERNAL, SCIDB_LE_MISMATCHED_BOUNDARIES);

    if (isEmpty() || other.isEmpty())
    {
        return createEmpty(_startCoords.size());
    }

    Coordinates start;
    for (size_t i =0; i<_startCoords.size(); i++)
    {
        Coordinate myStart = _startCoords[i];
        Coordinate otherStart = other._startCoords[i];
        start.push_back( myStart < otherStart ? otherStart : myStart );
    }

    Coordinates end;
    for (size_t i =0; i<_endCoords.size(); i++)
    {
        Coordinate myEnd = _endCoords[i];
        Coordinate otherEnd = other._endCoords[i];
        end.push_back( myEnd > otherEnd ? otherEnd : myEnd);
    }

    double intersectionCells = static_cast<double>(getNumCells(start,end));

    double resultDensity = 1.0;
    if (intersectionCells > 0)
    {
        double myCells = static_cast<double>(getNumCells());
        double otherCells = static_cast<double>(other.getNumCells());
        double maxMyDensity = min( _density * myCells / intersectionCells, 1.0 );
        double maxOtherDensity = min ( other._density * otherCells / intersectionCells, 1.0);
        resultDensity = min (maxMyDensity, maxOtherDensity);
    }

    return PhysicalBoundaries(start,end, resultDensity);
}

PhysicalBoundaries PhysicalBoundaries::unionWith (PhysicalBoundaries const& other) const
{
    if (_startCoords.size() != other._startCoords.size()
            || _startCoords.size() != other._endCoords.size())
        throw SYSTEM_EXCEPTION(SCIDB_SE_INTERNAL, SCIDB_LE_MISMATCHED_BOUNDARIES);

    if (isEmpty())
    {
        return other;
    }

    else if (other.isEmpty())
    {
        return *this;
    }

    Coordinates start;
    for (size_t i =0; i<_startCoords.size(); i++)
    {
        Coordinate myStart = _startCoords[i];
        Coordinate otherStart = other._startCoords[i];
        start.push_back( myStart > otherStart ? otherStart : myStart );
    }

    Coordinates end;
    for (size_t i =0; i<_endCoords.size(); i++)
    {
        Coordinate myEnd = _endCoords[i];
        Coordinate otherEnd = other._endCoords[i];
        end.push_back( myEnd < otherEnd ? otherEnd : myEnd);
    }

    double myCells = static_cast<double>(getNumCells());
    double otherCells = static_cast<double>(other.getNumCells());
    double resultCells = static_cast<double>(getNumCells(start, end));
    double maxDensity = min ( (myCells * _density + otherCells * other._density ) / resultCells, 1.0);

    return PhysicalBoundaries(start,end, maxDensity);
}

PhysicalBoundaries PhysicalBoundaries::crossWith (PhysicalBoundaries const& other) const
{
    if (isEmpty() || other.isEmpty())
    {
        return createEmpty(_startCoords.size()+other._startCoords.size());
    }

    Coordinates start, end;
    for (size_t i=0; i<_startCoords.size(); i++)
    {
        start.push_back(_startCoords[i]);
        end.push_back(_endCoords[i]);
    }
    for (size_t i=0; i<other.getStartCoords().size(); i++)
    {
        start.push_back(other.getStartCoords()[i]);
        end.push_back(other.getEndCoords()[i]);
    }

    return PhysicalBoundaries(start,end, _density * other._density);
}

namespace {
    bool hasMaxStar(Coordinates const& coords)
    {
        for (Coordinate const& c : coords) {
            if (CoordinateBounds::isMaxStar(c)) {
                return true;
            }
        }
        return false;
    }
}

PhysicalBoundaries PhysicalBoundaries::reshape(Dimensions const& oldDims, Dimensions const& newDims) const
{
    if (isEmpty())
    {
        return createEmpty(newDims.size());
    }

    // If *this* bounding box is already unbounded (hasMaxStar), don't complain if it won't fit into
    // the new shape.  That wouldn't be surprising.
    ASSERT_EXCEPTION(!hasMaxStar(_startCoords), "Starting coordinate somehow got set to '*'");
    const bool strict = !hasMaxStar(_endCoords);

    Coordinates start = reshapeCoordinates(_startCoords, oldDims, newDims, strict);
    Coordinates end = reshapeCoordinates(_endCoords, oldDims, newDims, strict);

    if (newDims.size() > oldDims.size())
    {
        bool dimensionFull = false;

        for (size_t i = 0; i < start.size(); i++)
        {
            if (dimensionFull)
            {
                if (newDims[i].getStartMin() <= CoordinateBounds::getMin()
                    || newDims[i].getEndMax() >= CoordinateBounds::getMax())
                {
                    throw SYSTEM_EXCEPTION(SCIDB_SE_INTERNAL, SCIDB_LE_CANT_CREATE_BOUNDARIES_FROM_INFINITE_ARRAY);
                }

                start[i] = newDims[i].getStartMin();
                end[i] = newDims[i].getEndMax();
            }
            else if(end[i] > start[i])
            {
                dimensionFull = true;
            }
        }
    }

    // Cell counts will be wildly wrong when bounding coordinates use '*' (MAX_COORDINATE), so use
    // std::min() to ensure that the density computation stays sane, in range [0.0 .. 1.0].
    double startingCells = static_cast<double>(getNumCells());
    double resultCells = static_cast<double>(getNumCells(start, end));
    double newDensity = std::min(1.0, _density * startingCells / resultCells);
    return PhysicalBoundaries(start, end, newDensity);
}

std::shared_ptr<SharedBuffer> PhysicalBoundaries::serialize() const
{
    size_t totalSize = sizeof(size_t) + sizeof(double) + _startCoords.size()*sizeof(Coordinate) + _endCoords.size()*sizeof(Coordinate);
    MemoryBuffer* buf = new MemoryBuffer(NULL, totalSize);

    auto sizePtr = reinterpret_cast<size_t*>(buf->getWriteData());
    auto startingPoint = sizePtr;
    *sizePtr = _startCoords.size();
    sizePtr++;

    double* densityPtr = (double*) sizePtr;
    *densityPtr = _density;
    densityPtr++;

    Coordinate* coordPtr = (Coordinate*) densityPtr;

    for(size_t i = 0; i<_startCoords.size(); i++)
    {
        *coordPtr = _startCoords[i];
        coordPtr++;
    }

    for(size_t i = 0; i<_endCoords.size(); i++)
    {
        *coordPtr = _endCoords[i];
        coordPtr++;
    }

    SCIDB_ASSERT(reinterpret_cast<char*>(coordPtr) == reinterpret_cast<char*>(startingPoint) + buf->getSize());
    return std::shared_ptr<SharedBuffer> (buf);
}

PhysicalBoundaries PhysicalBoundaries::deSerialize(std::shared_ptr<SharedBuffer> const& buf)
{
    auto numCoordsPtr = reinterpret_cast<const size_t*>(buf->getConstData());
    size_t numCoords = *numCoordsPtr;
    numCoordsPtr++;

    auto densityPtr = reinterpret_cast<const double*>(buf->getConstData());
    double density = *densityPtr;
    densityPtr++;

    Coordinates start, end;
    Coordinate* coordPtr = (Coordinate*) densityPtr;

    for(size_t i =0; i<numCoords; i++)
    {
        start.push_back(*coordPtr);
        coordPtr++;
    }

    for(size_t i =0; i<numCoords; i++)
    {
        end.push_back(*coordPtr);
        coordPtr++;
    }

    return PhysicalBoundaries(start,end,density);
}


uint32_t PhysicalBoundaries::getCellSizeBytes(const Attributes& attrs )
{
    uint32_t totalBitSize = 0;
    Config* cfg = Config::getInstance();

    for (size_t i = 0; i < attrs.size(); i++)
    {
        const AttributeDesc attr = attrs[i];
        Type cellType = TypeLibrary::getType(attr.getType());
        uint32_t bitSize = cellType.bitSize();

        if (bitSize == 0)
        {
            bitSize =  cfg->getOption<int>(CONFIG_STRING_SIZE_ESTIMATION) * 8;
        }
        if (attr.isNullable())
        {
            bitSize += 1;
        }
        totalBitSize += bitSize;
    }
    return (totalBitSize + 7)/8;
}

double PhysicalBoundaries::getSizeEstimateBytes(const ArrayDesc& schema) const
{
    uint64_t numCells = getNumCells();
    size_t numDimensions = schema.getDimensions().size();
    size_t numAttributes = schema.getAttributes().size();

    size_t cellSize = getCellSizeBytes(schema.getAttributes());

    //we assume that every cell is part of sparse chunk
    cellSize += numAttributes * (numDimensions  * sizeof(Coordinate) + sizeof(int));
    double size = static_cast<double>(numCells) * static_cast<double>(cellSize) ;

    // Assume all chunks are sparse and add header (if we can).
    uint64_t numChunks = 0;
    try {
        numChunks = getNumChunks(schema.getDimensions());
    } catch (UnknownChunkIntervalException&) {
        numChunks = 0;
    }
    size += static_cast<double>(numChunks) *
        static_cast<double>(numAttributes);

    return size * _density;
}

/**
 * Broadcast a "sync" message and wait for an acknowledgement from every instance.
 */
void sync(NetworkManager* networkManager, const std::shared_ptr<Query>& query, uint64_t instanceCount, bool isSendLocal=false)
{
    std::shared_ptr<MessageDesc> msg = std::make_shared<MessageDesc>(mtSyncRequest);
    std::shared_ptr<scidb_msg::DummyQuery> record = msg->getRecord<scidb_msg::DummyQuery>();
    msg->setQueryID(query->getQueryID());

    LOG4CXX_DEBUG(logger, "sync: Broadcast");
    networkManager->broadcastLogical(msg);
    if (isSendLocal) {
         networkManager->sendLocal(query, msg);
        ++instanceCount;
    }

    LOG4CXX_DEBUG(logger, "sync: Waiting for " << instanceCount-1 << " others");
    Semaphore::ErrorChecker ec = bind(&Query::validate, query);
    query->syncSG.enter(instanceCount - 1, ec, PTW_SEM_SG_RCV);

    LOG4CXX_DEBUG(logger, "sync: done.")
}


void barrier(uint64_t barrierId,
             NetworkManager* networkManager,
             const std::shared_ptr<Query>& query,
             uint64_t instanceCount)
{
    std::shared_ptr<MessageDesc> barrierMsg = std::make_shared<MessageDesc>(mtBarrier);
    std::shared_ptr<scidb_msg::DummyQuery> barrierRecord = barrierMsg->getRecord<scidb_msg::DummyQuery>();
    barrierMsg->setQueryID(query->getQueryID());
    barrierRecord->set_payload_id(barrierId);

    LOG4CXX_DEBUG(logger, "barrier: Broadcast");
    networkManager->broadcastLogical(barrierMsg);

    LOG4CXX_DEBUG(logger, "barrier: Waiting waiting for " << instanceCount - 1 << " others");
    Semaphore::ErrorChecker ec = bind(&Query::validate, query);
    query->semSG[barrierId].enter(instanceCount - 1, ec, PTW_SEM_BAR_SG);

    LOG4CXX_DEBUG(logger, "barrier: done.")
}

/**
 * This can be used as a barrier mechanism across the cluster in a blocking/materializing operator
 * @note: redistributeXXX() uses the same mechanism.
 * @todo: make a method directly on query?
 * @note: These methods on operator could remain (if needed).
 *        They would just be wrappers then.
 */
void syncBarrier(uint64_t barrierId, const std::shared_ptr<Query>& query)
{
    LOG4CXX_DEBUG(logger, "syncBarrier: barrierId = " << barrierId);
    SCIDB_ASSERT(query);
    NetworkManager* networkManager = NetworkManager::getInstance();
    SCIDB_ASSERT(networkManager);
    const uint64_t instanceCount = query->getInstancesCount();
    SCIDB_ASSERT(instanceCount > 0);
    barrier(barrierId%MAX_BARRIERS, networkManager, query, instanceCount);
    LOG4CXX_DEBUG(logger, "syncBarrier: returning");
}

void syncSG(const std::shared_ptr<Query>& query)
{
    SCIDB_ASSERT(query);
    LOG4CXX_DEBUG(logger, "syncSG: queryID="<<query->getQueryID());
    NetworkManager* networkManager = NetworkManager::getInstance();
    SCIDB_ASSERT(networkManager);
    const uint64_t instanceCount = query->getInstancesCount();
    SCIDB_ASSERT(instanceCount > 0);
    sync(networkManager, query, instanceCount, true);
    LOG4CXX_DEBUG(logger, "syncSG: returning");
}

AggregatePtr resolveAggregate(std::shared_ptr <OperatorParamAggregateCall>const& aggregateCall,
                              Attributes const& inputAttributes,
                              AttributeID* inputAttributeID,
                              string* outputName)
{
    const std::shared_ptr<OperatorParam> &acParam = aggregateCall->getInputAttribute();

    try
    {
        if (PARAM_ASTERISK == acParam->getParamType())
        {
            AggregatePtr agg = AggregateLibrary::getInstance()->createAggregate( aggregateCall->getAggregateName(), TypeLibrary::getType(TID_VOID));

            if (inputAttributeID)
            {
                *inputAttributeID = INVALID_ATTRIBUTE_ID;
            }
            if (outputName)
            {
                *outputName = aggregateCall->getAlias().size() ? aggregateCall->getAlias() : agg->getName();
            }
            return agg;
        }
        else if (PARAM_ATTRIBUTE_REF == acParam->getParamType())
        {
            const std::shared_ptr<OperatorParamAttributeReference> &ref = (const std::shared_ptr<OperatorParamAttributeReference>&) acParam;

            AttributeDesc const& inputAttr = inputAttributes[ref->getObjectNo()];
            Type const& inputType = TypeLibrary::getType(inputAttr.getType());
            AggregatePtr agg = AggregateLibrary::getInstance()->createAggregate( aggregateCall->getAggregateName(), inputType);

            if (inputAttributeID)
            {
                *inputAttributeID = inputAttr.getId();
            }
            if (outputName)
            {
                *outputName = aggregateCall->getAlias().size() ? aggregateCall->getAlias() : inputAttr.getName() + "_" + agg->getName();
            }
            return agg;
        }
        else
        {
            // All other cases must have been thrown already during translation.
            // (If assertions disabled, on to throw statement at end of function.)
            SCIDB_ASSERT(0);
        }
    }
    catch(const UserException &e)
    {
        if (SCIDB_LE_AGGREGATE_NOT_FOUND == e.getLongErrorCode())
        {
            throw CONV_TO_USER_QUERY_EXCEPTION(e, acParam->getParsingContext());
        }

        throw;
    }

    SCIDB_ASSERT(false);
    throw SYSTEM_EXCEPTION(SCIDB_SE_INTERNAL, SCIDB_LE_UNREACHABLE_CODE)
        << __FUNCTION__;
    return AggregatePtr();
}

void addAggregatedAttribute (
        std::shared_ptr <OperatorParamAggregateCall>const& aggregateCall,
        ArrayDesc const& inputDesc,
        ArrayDesc& outputDesc,
        bool operatorDoesAggregationInOrder)
{
    string outputName;

    AggregatePtr agg = resolveAggregate(aggregateCall, inputDesc.getAttributes(), 0, &outputName);

    if ( !operatorDoesAggregationInOrder && agg->isOrderSensitive() ) {
        throw USER_EXCEPTION(SCIDB_SE_OPERATOR, SCIDB_LE_AGGREGATION_ORDER_MISMATCH) << agg->getName();
    }

    outputDesc.addAttribute(
        AttributeDesc(
            safe_static_cast<AttributeID>(outputDesc.getAttributes().size()),
            outputName,
            agg->getResultType().typeId(),
            AttributeDesc::IS_NULLABLE,
            CompressorType::NONE));
}

void PhysicalBoundaries::updateFromChunk(ConstChunk const* chunk, bool chunkShapeOnly)
{
    size_t nDims = _startCoords.size();
    if (chunk == NULL)
    {   return; }

    //chunk iteration is expensive - only perform if needed
    Coordinates const& chunkFirstPos = chunk->getFirstPosition(false), chunkLastPos = chunk->getLastPosition(false);
    bool updateLowBound = false, updateHiBound = false;
    for (size_t i = 0; i < nDims; i++)
    {
       if (chunkFirstPos[i] < _startCoords[i])
       {
           if ( chunkShapeOnly )
           {    _startCoords[i] = chunkFirstPos[i]; }
           else
           {   updateLowBound = true; }
       }

       if (chunkLastPos[i] > _endCoords[i])
       {
           if ( chunkShapeOnly)
           {    _endCoords[i] = chunkLastPos[i]; }
           else
           {    updateHiBound = true; }
       }
    }

    //The chunk is inside the box and cannot expand the bounds. Early exit.
    if (!updateLowBound && !updateHiBound)
    {   return; }

    //TODO: there is a further optimization opportunity here. The given chunk *should* always be a bitmap
    //chunk. Once we verify that, we can iterate over bitmap segments and compute coordinates. Committing
    //this due to timing constraints - should revisit and optimize this when possible.
    std::shared_ptr<ConstChunkIterator> citer = chunk->materialize()->getConstIterator();
    while (!citer->end() && (updateLowBound || updateHiBound))
    {
        Coordinates const& pos = citer->getPosition();
        bool updated = false;
        for (size_t j = 0; j < nDims; j++)
        {
            if (updateHiBound && pos[j] > _endCoords[j])
            {
                _endCoords[j] = pos[j];
                updated = true;
            }
            if (updateLowBound && pos[j] < _startCoords[j])
            {
                _startCoords[j] = pos[j];
                updated = true;
            }
        }
        if(updated) //it's likely that no further update can come from this chunk
        {
            if(updateHiBound)
            {
                size_t k=0;
                while (k<nDims && _endCoords[k]>=chunkLastPos[k])
                { k++; }
                if (k==nDims) //no more useful data for hi bound could come from this chunk!
                {   updateHiBound=false; }
            }
            if(updateLowBound)
            {
                size_t k=0;
                while (k<nDims && _startCoords[k]<=chunkFirstPos[k])
                { k++; }
                if (k==nDims) //no more useful data for lo bound could come from this chunk!
                {   updateLowBound=false; }
            }
        }
        ++(*citer);
    }
}

PhysicalBoundaries PhysicalBoundaries::trimToDims(Dimensions const& dims) const
{
    SCIDB_ASSERT(_startCoords.size() == dims.size());
    size_t nDims = dims.size();

    Coordinates resStart(nDims);
    Coordinates resEnd(nDims);

    for(size_t i=0; i<nDims; i++)
    {
        resStart[i] = std::max<Coordinate> (dims[i].getStartMin(), _startCoords[i]);
        resEnd[i] = std::min<Coordinate> (dims[i].getEndMax(), _endCoords[i]);
    }

    return PhysicalBoundaries(resStart, resEnd, _density);
}

PhysicalBoundaries PhysicalBoundaries::createFromChunkList(std::shared_ptr<Array>& inputArray,
                                                           const std::set<Coordinates, CoordinatesLess>& chunkCoordinates)
{
    ASSERT_EXCEPTION((inputArray->getSupportedAccess() == Array::RANDOM),
                     "PhysicalBoundaries::createFromChunkList: ");

    const ArrayDesc& arrayDesc = inputArray->getArrayDesc();
    const size_t nDims = arrayDesc.getDimensions().size();

    if (chunkCoordinates.empty()) {
        return PhysicalBoundaries::createEmpty(nDims);
    }

    typedef set<Coordinates, CoordinatesLess> ChunkCoordinates;
    typedef std::unordered_set<const Coordinates*> CoordHashSet;
    CoordHashSet chunksToExamine;

    if (nDims>1) {

        typedef std::vector<std::pair<int64_t, std::deque<const Coordinates*> > > ChunkListByExtremeDim;
        ChunkListByExtremeDim chunksWithMinDim(nDims,make_pair(CoordinateBounds::getMax(),std::deque<const Coordinates*>()));
        SCIDB_ASSERT(nDims == chunksWithMinDim.size());
        ChunkListByExtremeDim chunksWithMaxDim(nDims,make_pair(CoordinateBounds::getMin(),std::deque<const Coordinates*>()));
        SCIDB_ASSERT(nDims == chunksWithMaxDim.size());

        // find all boundary chunks
        for (ChunkCoordinates::const_iterator iter = chunkCoordinates.begin();
             iter != chunkCoordinates.end(); ++iter) {
            const Coordinates& chunkCoords = *iter;
            SCIDB_ASSERT(chunkCoords.size() == nDims);
            for (size_t d=0; d < nDims; ++d) {
                pair<int64_t, std::deque<const Coordinates*> >& minChunks = chunksWithMinDim[d];
                if (chunkCoords[d] < minChunks.first) {
                    minChunks.second.clear();
                    minChunks.first = chunkCoords[d];
                    minChunks.second.push_back(&chunkCoords);
                }  else if (chunkCoords[d] == minChunks.first) {
                    minChunks.second.push_back(&chunkCoords);
                } else { }

                pair<int64_t, std::deque<const Coordinates*> >& maxChunks = chunksWithMaxDim[d];
                if (chunkCoords[d] > maxChunks.first) {
                    maxChunks.second.clear();
                    maxChunks.first = chunkCoords[d];
                    maxChunks.second.push_back(&chunkCoords);
                }  else if (chunkCoords[d] == maxChunks.first) {
                    maxChunks.second.push_back(&chunkCoords);
                } else { }
            }
        }

        // throw away duplicates
        for (size_t d=0; d < nDims; ++d) {
            std::deque<const Coordinates*>& minList = chunksWithMinDim[d].second;
            for (std::deque<const Coordinates*>::const_iterator iter = minList.begin();
                 iter != minList.end(); ++iter) {
                chunksToExamine.insert(*iter).second;
            }

            std::deque<const Coordinates*>& maxList = chunksWithMaxDim[d].second;
            for (std::deque<const Coordinates*>::const_iterator iter = maxList.begin();
                 iter != maxList.end(); ++iter) {
                chunksToExamine.insert(*iter).second;
            }
        }
    } else {
        SCIDB_ASSERT(nDims == 1);
        const Coordinates& firstCoord = *chunkCoordinates.begin();
        chunksToExamine.insert(&firstCoord);
        const Coordinates& lastCoord = *(--chunkCoordinates.end());
        chunksToExamine.insert(&lastCoord);
    }

    // update bounds using the boundary chunks
    PhysicalBoundaries bounds = PhysicalBoundaries::createEmpty(nDims);

    const AttributeID attr = safe_static_cast<AttributeID>(arrayDesc.getAttributes().size() - 1);
    const bool isNoEmptyTag = (arrayDesc.getEmptyBitmapAttribute() == NULL);
    std::shared_ptr<ConstArrayIterator> arrayIter = inputArray->getConstIterator(attr);
    for (CoordHashSet::const_iterator iter = chunksToExamine.begin();
         iter != chunksToExamine.end(); ++iter) {
        const Coordinates* chunkCoords = *iter;
        bool rc = arrayIter->setPosition(*chunkCoords);
        SCIDB_ASSERT(rc);
        const ConstChunk& chunk = arrayIter->getChunk();
        bounds.updateFromChunk(&chunk, isNoEmptyTag);
    }
    return bounds;
}

void BaseLogicalOperatorFactory::registerFactory()
{
    OperatorLibrary::getInstance()->addLogicalOperatorFactory(this);
}

void BasePhysicalOperatorFactory::registerFactory()
{
    OperatorLibrary::getInstance()->addPhysicalOperatorFactory(this);
}

VersionID OperatorParamArrayReference::getVersion() const
{
    return _version;
}

LogicalOperator::LogicalOperator(const std::string& logicalName, const std::string& aliasName)
:   _logicalName(logicalName),
    _hasSchema(false),
    _aliasName(aliasName)
{
}

void LogicalOperator::setSchema(const ArrayDesc& newSchema)
{
    if (hasSchema()) {
        LOG4CXX_WARN(logger, "LogicalOperator::setSchema() updating existing _schema: " << getSchema()
                             << " to: "  << newSchema); 
    }

    _schema = newSchema;
    _hasSchema = true;

    // historical behavior: a non-empty _aliasName on this overrides the name from newSchema
    if (_aliasName.size() != 0) {
        _schema.setName(_aliasName); 
    }
}

bool LogicalOperator::hasSchema() const
{
    return _hasSchema;
}

const ArrayDesc& LogicalOperator::getSchema() const
{
    SCIDB_ASSERT(hasSchema());
    return _schema;
}

void LogicalOperator::setAliasName(const std::string &alias)
{
    // TODO: consider changing callers so we can assert !alias.empty()
    // TODO: consider changing callers so we can assert _aliasName.empty()

    if(!_aliasName.empty() && _aliasName != alias) {
        LOG4CXX_WARN(logger, "LogicalOperator::setAliasName changing from" << _aliasName << " to " << alias);
    }

    _aliasName = alias;
}

void LogicalOperator::inferAccess(std::shared_ptr<Query>& query)
{
    for (Parameter& p : _parameters) {
        _requestLeastAccess(p, query);
    }
    for (auto& kwPair : _kwParameters) {
        _requestLeastAccess(kwPair.second, query);
    }
}

void LogicalOperator::_requestLeastAccess(Parameter const& param,
                                          shared_ptr<Query>& query)
{
    string arrayNameOrig;
    bool isSchema = false;
    if (param->getParamType() == PARAM_ARRAY_REF) {
        arrayNameOrig = ((std::shared_ptr<OperatorParamReference>&)param)->getObjectName();
    } else if (param->getParamType() == PARAM_SCHEMA) {
        arrayNameOrig = ((std::shared_ptr<OperatorParamSchema>&)param)->getSchema().getQualifiedArrayName();
        isSchema = true;
    }
    if (arrayNameOrig.empty()) {
        return;
    }

    std::string arrayName;
    std::string namespaceName;
    query->getNamespaceArrayNames(arrayNameOrig, namespaceName, arrayName);
    const string baseName = ArrayDesc::makeUnversionedName(arrayName);

    std::shared_ptr<LockDesc> lock(
        make_shared<LockDesc>(
            namespaceName,
            baseName,
            query->getQueryID(),
            Cluster::getInstance()->getLocalInstanceId(),
            LockDesc::COORD,
            LockDesc::RD));
    query->requestLock(lock);

    // Only grant automatic namespace read rights if this array
    // name is used as a schema parameter.  Here's why:
    //
    // Unlike catalog locks, which are hierarchical (RD < WR <
    // XCL), there is no such ordering for namespace rights.  An
    // operator having namespace write access (store, insert) does
    // @em not automatically imply namespace read access.  Derived
    // classes must explicitly ask for (only) the namespace rights
    // they need.
    //
    // Often, operators call this base class method before
    // upgrading the catalog read lock(s) acquired above, so they
    // don't -really- need read access to the namespace.  That's
    // inaccurate and wasteful, oh well.  But if an array is being
    // used just for its schema, then definitely you need read
    // access to -that- namespace.  So here we request it.
    //
    if (isSchema) {
        query->getRights()->upsert(rbac::ET_NAMESPACE, namespaceName,
                                   rbac::P_NS_READ);
    }
}

bool LogicalOperator::hasRightsForAllNamespaces(shared_ptr<Query>& query)
{
    // This check is not foolproof: the current operator may have
    // failed to specify rights for namespace "ns1", but another
    // operator in the tree already did so, masking this operator's
    // failure.  Oh well, we do the best we can.

    int missing = 0;
    for (Parameter& p : _parameters) {
        if (!_hasNamespaceRights(p, query)) {
            ++missing;
        }
    }
    for (auto& kwPair : _kwParameters) {
        if (!_hasNamespaceRights(kwPair.second, query)) {
            ++missing;
        }
    }
    return missing == 0;
}

bool LogicalOperator::_hasNamespaceRights(Parameter const& param,
                                          shared_ptr<Query> const& query)
{
    string arrayNameOrig;
    if (param->getParamType() == PARAM_ARRAY_REF) {
        arrayNameOrig = ((std::shared_ptr<OperatorParamReference>&)param)->getObjectName();
    } else if (param->getParamType() == PARAM_SCHEMA) {
        arrayNameOrig = ((std::shared_ptr<OperatorParamSchema>&)param)->getSchema().getQualifiedArrayName();
    }
    if (arrayNameOrig.empty()) {
        return true;
    }

    string ns, ary;
    query->getNamespaceArrayNames(arrayNameOrig, ns, ary);
    rbac::RightsMap const& rights = *query->getRights();
    auto pos = rights.find(rbac::EntityTag(rbac::ET_NAMESPACE, ns));
    if (pos == rights.end()) {
        LOG4CXX_ERROR(logger, "Operator " << getLogicalName()
                      << " did not acquire namespace rights for array "
                      << arrayNameOrig << " in namespace " << ns);
    } else {
        // Assert that at least one permission bit is set.
        SCIDB_ASSERT(pos->second);
    }
    return pos != rights.end();
}

InjectedErrorListener PhysicalOperator::_injectedErrorListener(InjectErrCode::OPERATOR);

std::shared_ptr<ThreadPool> PhysicalOperator::_globalThreadPoolForOperators;
std::shared_ptr<JobQueue> PhysicalOperator::_globalQueueForOperators;
Mutex PhysicalOperator::_mutexGlobalQueueForOperators;

std::shared_ptr<JobQueue> PhysicalOperator::getGlobalQueueForOperators()
{
    ScopedMutexLock cs(_mutexGlobalQueueForOperators, PTW_SML_OPERATOR);
    if (!_globalThreadPoolForOperators) {
        _globalQueueForOperators = std::shared_ptr<JobQueue>(new JobQueue("GlobalJobQueueForOperators"));
        _globalThreadPoolForOperators = std::shared_ptr<ThreadPool>(
                new ThreadPool(Config::getInstance()->getOption<int>(CONFIG_RESULT_PREFETCH_THREADS),
                               _globalQueueForOperators,
                               "globalOperatorThreadPool"));
        _globalThreadPoolForOperators->start();
    }
    return _globalQueueForOperators;
}


void PhysicalOperator::checkOrUpdateIntervals(ArrayDesc& checkMe, std::shared_ptr<Array>& input)
{
    checkOrUpdateIntervals(checkMe, input->getArrayDesc().getDimensions());
}

void PhysicalOperator::checkOrUpdateIntervals(ArrayDesc& checkMe, Dimensions const& inDims)
{
    // If the input schema was autochunked, we must now fix it up with actual intervals on *all* the
    // instances.  If *not* autochunked... well, we need to scan the dimensions anyway to find that
    // out... so we might as well do this inexpensive check in that case as well.

    Dimensions& cDims = checkMe.getDimensions();
    const size_t N_DIMS = cDims.size();
    SCIDB_ASSERT(inDims.size() == N_DIMS);

    for (size_t i = 0; i < N_DIMS; ++i) {
        int64_t hardInterval = inDims[i].getChunkInterval();
        int64_t softInterval = cDims[i].getRawChunkInterval();
        if (! cDims[i].isIntervalResolved()) {
            cDims[i].setChunkInterval(hardInterval);
        } else if (softInterval != hardInterval) {
            // By throwing here, we'll have wasted our earlier call to catalog->getNextArrayId(), and
            // burned an array id.  Oh well.  (This is re. calls from PhysicalUpdate.)

            // With autochunking, it's the optimizer (or something that should eventually be in the
            // optimizer) that's responsible for chunk size mismatches getting resolved.  So we use
            // that "short" error code when doing the late re-checking of the intervals.
            throw USER_EXCEPTION(SCIDB_SE_OPTIMIZER, SCIDB_LE_DIMENSIONS_DONT_MATCH)
                << inDims[i] << cDims[i];
        }
    }
}


void PhysicalOperator::repartByLeftmost(
    vector<ArrayDesc> const& inputSchemas,
    vector<ArrayDesc const*>& modifiedPtrs) const
{
    const size_t N_SCHEMAS = inputSchemas.size();
    SCIDB_ASSERT(N_SCHEMAS > 1); // ... else you are calling the wrong canned implementation.
    SCIDB_ASSERT(N_SCHEMAS == modifiedPtrs.size());
    // NOTE:  N_SCHEMAS may (someday) be  > 2

    // Find leftmost *non-autochunked* schema, use *that* as the exemplar for
    // the following aspects of dimensions: chunkInterval and chunkOverlap.  The
    // names, start indices, and end indices of dimensions are preserved for each
    // of the input schemas.
    size_t exemplarIndex = 0;
    for (ArrayDesc const& schema : inputSchemas) {
        if (schema.isAutochunked()) {
                ++exemplarIndex;
        }
        else {
            break;
        }
    }
    // At least one of the input schemas must be non-autochunked
    if (exemplarIndex == N_SCHEMAS) {
        throw USER_EXCEPTION(SCIDB_SE_OPERATOR, SCIDB_LE_ALL_INPUTS_AUTOCHUNKED)
            << getLogicalName();
    }

    // We don't expect to be called twice, but that may change later on:
    //  wipe any previous result.
    _redimRepartSchemas.clear();

    // Do not repartition leftmost (non-autochunked) input array.
    modifiedPtrs[exemplarIndex] = nullptr;

    ArrayDesc const& exemplarSchema = inputSchemas[exemplarIndex];
    const size_t     nDimensions    = exemplarSchema.getDimensions().size();
    // Check the other input schemas, adding a desired schema to modifiedPtrs
    // for each input that needs to have an automatic repart operator added.
    for (size_t iSchema = 0; iSchema < N_SCHEMAS; ++iSchema)
    {
        if (iSchema == exemplarIndex) {
            // This is the exemplar Schema. There is no need to insert a
            // repart/redimension to make the partitioning match itself.
            continue;
        }

        ArrayDesc const& targetSchema = inputSchemas[iSchema];
        if (exemplarSchema.samePartitioning(targetSchema)) {
            // Already has correct chunkSize and overlap, do nothing.
            modifiedPtrs[iSchema] = nullptr;
        }
        else {
            // Clone this schema and adjust dimensions according to exemplar.
            Dimensions const& exemplarDimensions = exemplarSchema.getDimensions();
            // maintain the names of the dimensions by starting with the
            // original dimension information from the input schema
            Dimensions targetDimensions(targetSchema.getDimensions());
            SCIDB_ASSERT(targetDimensions.size() == nDimensions);
            for (size_t iDim = 0; iDim < nDimensions; ++iDim)
            {
                DimensionDesc const & exemplarDim = exemplarDimensions[iDim];
                DimensionDesc       & targetDim   = targetDimensions[iDim];

                targetDim.setChunkInterval(exemplarDim.getChunkInterval());

                // Take smallest overlap since we can't (easily) conjure up cells that aren't there.
                targetDim.setChunkOverlap(min(exemplarDim.getChunkOverlap(),
                                              targetDim.getChunkOverlap()));
            }

            _redimRepartSchemas.push_back(make_shared<ArrayDesc>(
                                              targetSchema.getName(),
                                              targetSchema.getAttributes(),
                                              targetDimensions,
                                              targetSchema.getDistribution(),
                                              targetSchema.getResidency()));
            modifiedPtrs[iSchema] = _redimRepartSchemas.back().get();
        }
    }
    if (_redimRepartSchemas.empty()) {
        // Assertions elsewhere hate an all-NULLs vector here.
        modifiedPtrs.clear();
    }
}


void PhysicalOperator::repartForStoreOrInsert(
    vector<ArrayDesc> const& inputSchemas,
    vector<ArrayDesc const*>& modifiedPtrs) const
{
    SCIDB_ASSERT(inputSchemas.size() == 1);
    SCIDB_ASSERT(modifiedPtrs.size() == 1);

    // If schema matches input array, no problem.  Loose comparison allows
    // unspecified dimensions to match.
    if (_schema.sameLoosePartitioning(inputSchemas[0])) {
        modifiedPtrs.clear();
        return;
    }

    // A repart is needed.  Start with a copy of _schema, the stored-into
    // target array, because the input must be repartitioned to match it.
    _redimRepartSchemas.clear();
    _redimRepartSchemas.push_back(make_shared<ArrayDesc>(_schema));

    // If any target dimension's chunk interval is unresolved, take it from
    // the input as-is.
    Dimensions const& srcDims = inputSchemas[0].getDimensions();
    Dimensions& dstDims = _redimRepartSchemas.back()->getDimensions();
    size_t const N_DIMS = srcDims.size();
    SCIDB_ASSERT(dstDims.size() == N_DIMS);
    for (size_t d = 0; d < N_DIMS; ++d) {
        if (!dstDims[d].isIntervalResolved()) {
            dstDims[d].setRawChunkInterval(srcDims[d].getRawChunkInterval());
        }
    }

    // Request the repartition.
    modifiedPtrs[0] = _redimRepartSchemas[0].get();
}


void
PhysicalOperator::assertLastVersion(std::string const& nsName,
                                    std::string const& arrayName,
                                    bool arrayExists,
                                    ArrayDesc const& desc)
{
    if (isDebug()) {
        ArrayDesc tmpDesc;
        SystemCatalog::GetArrayDescArgs args;
        args.result = &tmpDesc;
        args.nsName = nsName;
        args.arrayName = arrayName;
        args.throwIfNotFound = false;
        bool found = SystemCatalog::getInstance()->getArrayDesc(args);

        SCIDB_ASSERT(arrayExists == found);

        if (arrayExists) {
            SCIDB_ASSERT(desc == tmpDesc);
        }
    }
}

void
PhysicalOperator::assertConsistency(ArrayDesc const& inputSchema,
                                    RedistributeContext const& inputDistribution)
{
   if (isDebug()) {
       SCIDB_ASSERT(inputSchema.getResidency()->isEqual(inputDistribution.getArrayResidency()));
       if (inputSchema.getId() > 0 && inputDistribution.getArrayDistribution()->getPartitioningSchema() == psUndefined) {
           // in degraded mode scan() reports psUndefined but
           // it has to preserve the original distribution for SMGR to make sense of the chunks
       } else {
           SCIDB_ASSERT(inputSchema.getDistribution()->checkCompatibility(inputDistribution.getArrayDistribution()));
       }
   }
}

std::string paramToString(Parameter const& p)
{
    // Try logical first...
    auto const lexp = dynamic_cast<OperatorParamLogicalExpression const*>(p.get());
    if (lexp) {
        return evaluate(lexp->getExpression(), TID_STRING).getString();
    }
    auto const pexp = dynamic_cast<OperatorParamPhysicalExpression const *>(p.get());
    ASSERT_EXCEPTION(pexp, "Expression parameter required");
    return pexp->getExpression()->evaluate().getString();
}

} //namespace
